%%
%
%             DO NOT USE WITH SYMMETRIC POINT SETS.
%
% Given the same point set in two coordinate systems match the
% corresponding points. This may fail if the point sets are symmetrical, 
% the transformation relating the two point sets is not a rigid one or the 
% noise level is sufficiently high.
%
% @param pointsInLeft A 3xn matrix where the columns are the points in the
%                     first (left) coordinate system.
%
% @param pointsInRight A 3xn matrix where the columns are the points in the
%                      second (right) coordinate system.
%
% @return matchingPointsInRight A 3xn matrix where the columns are the points in the
%                      second (right) coordinate system. Point order is
%                      such that the i'th column of matchingPointsInRight corresponds 
%                      to column i in pointsInLeft.
%
function matchingPointsInRight = matchPointsBasedOnDistances(pointsInLeft, pointsInRight)
  
  UNIQUENESS_THRESHOLD = 1.0;
  n = size(pointsInLeft,2);
  
         %compute the distance between each point and all others and sort
         %them. we end with an n-1xn-1 matrix
  sortedDistancesInLeft = [];
  sortedDistancesInRight = [];
  for i=1:n
    p=pointsInLeft(:,i);     %compute distance between current point and all others
    subtractedData = [pointsInLeft(:,1:i-1),pointsInLeft(:,i+1:n)] - p(:,ones(1,n-1));
    sortedDistancesInLeft = [sortedDistancesInLeft; sort(sqrt(sum(subtractedData.^2,1)))];
    
    p=pointsInRight(:,i);     %compute distance between current point and all others
    subtractedData = [pointsInRight(:,1:i-1),pointsInRight(:,i+1:n)] - p(:,ones(1,n-1));
    sortedDistancesInRight = [sortedDistancesInRight; sort(sqrt(sum(subtractedData.^2,1)))];
  end
        %checking for uniqueness, more than one point has the same distances
        %to all others
  for i=1:n
   di = sortedDistancesInLeft(i,:);
   absSubtractedDistances = abs(sortedDistancesInLeft(i+1:n,:) - di(ones(1,n-i),:));
   for j=1:n-i
     if isempty(find(absSubtractedDistances(j,:)>UNIQUENESS_THRESHOLD))
       error('Distances between points are not unique (symmetry?).');
     end
   end
   di = sortedDistancesInRight(i,:);
   absSubtractedDistances = abs(sortedDistancesInRight(i+1:n,:) - di(ones(1,n-i),:));
   for j=1:n-i
     if isempty(find(absSubtractedDistances(j,:)>UNIQUENESS_THRESHOLD))
       error('Distances between points are not unique (symmetry?).');
     end
   end
  end
          %match points in right coordinate system to those in the left coordinate
          %system based on the minimal difference between distances
  indexes = [];          
  for i=1:n
    di = sortedDistancesInLeft(i,:);
    [val,index] = min(sum(abs(sortedDistancesInRight - di(ones(1,n),:)),2));
    indexes = [indexes, index];
  end
        %make sure the indexes are unique as the approach above does not
        %necessarily guarantee this
  if size(unique(indexes),2) ~= size(indexes,2)
    error('Non unique match.');
  end
  
  matchingPointsInRight = pointsInRight(:,indexes);
  