from __future__ import print_function

__author__ = 'Ziv Yaniv'

"""
This module contains functions associated with the task of estimating the
absolute orientation. Estimating the rigid transformation between two 
coordinate systems using a set of three or more corresponding points.
The module contains a method based on representing the rotation as a unit
quaternion due to B.K.P. Horn [1], and a method based on representing the rotation
as an orthonormal matrix due to S. Umeyama [2]. The former method is only applicable
in 3D and the latter in any dimension. From an accuracy standpoint they are
equivalent [3].

The relevant publications are:
[1] B.K.P. Horn, "Closed-form solution of absolute orientation using unit quaternions",
Journal of the Optical Society of America A, 1987.

[2] S. Umeyama, "Least-Squares Estimation of Transformation Parameters Between Two 
Point Patterns", IEEE Trans. Pattern Anal. Machine Intell, 1991.

[3] D. W. Eggert, A. Lorusso, R. B. Fisher, "Estimating 3-D rigid body 
transformations: a comparison of four major algorithms", Mach. Vis. Appl., 1997.
"""

import numpy as np
from scipy import linalg

def compute_pose_quaternion(points_in_left, points_in_right):
    """
    Compute the rigid transformation, T, such that T*pointsInLeft = pointsInRight.
    
    Args:
        points_in_left (list(tuple)): Set of points corresponding to points_in_right in a different coordinate system.
        points_in_right (list(tuple)): Set of points corresponding to points_in_left in a different coordinate system.
        
    Returns:
        R,t (numpy.ndarray, numpy.array): Rigid transformation that maps points_in_left onto points_in_right.
                                          R*points_in_left + t = points_in_right
    
    Notes: 
        This is an implementation of the quaternion based solution described
        in "Closed-form solution of absolute orientation using unit 
        quaternions", B.K.P. Horn, Journal of the Optical Society of America, 
        Vol. 4(4), pp 629--642, 1987.  
    """
    num_points = len(points_in_left)
    dim_points = len(points_in_left[0])
           #cursory check that the number of points is sufficient
    if num_points<dim_points:      
        raise ValueError('Number of points must be greater/equal {0}.'.format(dim_points))

             #construct matrices out of the two point sets for easy manipulation
    left_mat = np.array(points_in_left).T
    right_mat = np.array(points_in_right).T
   
    left_mean = left_mat.mean(1)
    right_mean = right_mat.mean(1)

    M = left_mat.dot(right_mat.T) - \
        np.outer(left_mean, right_mean)*num_points        
        
    delta = np.array([[M[1,2] - M[2,1]], [M[2,0] - M[0,2]], [M[0,1] - M[1,0]]])
    N = np.vstack((np.array([np.trace(M), delta[0], delta[1], delta[2]]), \
                   np.hstack((delta,M + M.T - np.trace(M)*np.eye(3)))))
    (eigenValues, eigenVectors) = linalg.eig(N)
            #quaternion we want is the eigenvector corresponding to the 
            #maximal eigenvalue (these are real valued as N is a real, symmetric 
            #matrix)
    R = _quaternion_to_matrix(eigenVectors[:,eigenValues.argmax()])  
    t = right_mean - R.dot(left_mean)
    return R, t

def compute_pose_matrix(points_in_left, points_in_right):
    '''
    Absolute orientation using a matrix to represent the rotation. Solution is due to
    S. Umeyama, "Least-Squares Estimation of Transformation Parameters 
    Between Two Point Patterns", IEEE Trans. Pattern Anal. Machine Intell., vol. 13(4): 376-380.
    
    This is a refinement of the method proposed by Arun, Huang and Blostein, ensuring that the 
    rotation matrix is indeed a rotation and not a reflection. 
    
    Args:
        points_in_left (list(tuple)): Set of points corresponding to points_in_right in a different coordinate system.
        points_in_right (list(tuple)): Set of points corresponding to points_in_left in a different coordinate system.
        
    Returns:
        R,t (numpy.ndarray, numpy.array): Rigid transformation that maps points_in_left onto points_in_right.
                                          R*points_in_left + t = points_in_right
    '''
    num_points = len(points_in_left)
    dim_points = len(points_in_left[0])
           #cursory check that the number of points is sufficient
    if num_points<dim_points:      
        raise ValueError('Number of points must be greater/equal {0}.'.format(dim_points))

             #construct matrices out of the two point sets for easy manipulation
    left_mat = np.array(points_in_left).T
    right_mat = np.array(points_in_right).T
     
         #center both data sets on the mean
    left_mean = left_mat.mean(1)
    right_mean = right_mat.mean(1)
    left_M = left_mat - np.tile(left_mean, (num_points, 1)).T     
    right_M = right_mat - np.tile(right_mean, (num_points, 1)).T     
    
    M = left_M.dot(right_M.T)               
    U,S,Vt = linalg.svd(M)
    V=Vt.T
         #V * diag(1,1,det(U*V)) * U' - diagonal matrix ensures that we have a rotation and not a reflection
    R = V.dot(np.diag((1,1,linalg.det(U.dot(V))))).dot(U.T) 
    t = right_mean - R.dot(left_mean) 
    return R,t

#return a matrix represetation of rotation from a unit quaternion representation
#the quaternion is assumed to have unit norm and q = [s,qx,qy,qz] 
def _quaternion_to_matrix(q):
   return np.array([[1-2*q[2]**2-2*q[3]**2,   2*q[1]*q[2]-2*q[0]*q[3], 2*q[1]*q[3]+2*q[0]*q[2]],
                    [2*q[1]*q[2]+2*q[0]*q[3], 1-2*q[1]**2-2*q[3]**2,   2*q[2]*q[3]-2*q[0]*q[1]],
                    [2*q[1]*q[3]-2*q[0]*q[2], 2*q[2]*q[3]+2*q[0]*q[1], 1-2*q[1]**2-2*q[2]**2]]) 
       

def create_fixed_axis_frameD(theta_x, theta_y, theta_z, dx, dy, dz):
    """
    Create a fixed axis frame from the given rotation angles and translation.
    Rotation angles are given in degrees.
    
    Args:
        theta_x (float): Rotation around x axis.
        theta_y (float): Rotation around y axis.
        theta_z (float): Rotation around z axis.
        dx (float): Translation along x axis.
        dy (float): Translation along y axis.
        dz (float): Translation along z axis.
    
    Returns:
        R,t (numpy.ndarray, numpy.array): Rigid transformation represented as
        a rotation matrix and translation.
    """
    return create_fixed_axis_frame(np.radians(theta_x), np.radians(theta_y),
                                 np.radians(theta_z), dx, dy, dz)


def create_fixed_axis_frame(theta_x, theta_y, theta_z, dx, dy, dz):
    """
    Create a fixed axis frame from the given rotation angles and translation.
    Rotation angles are given in radians.
    
    Args:
        theta_x (float): Rotation around x axis.
        theta_y (float): Rotation around y axis.
        theta_z (float): Rotation around z axis.
        dx (float): Translation along x axis.
        dy (float): Translation along y axis.
        dz (float): Translation along z axis.

    Returns:
        R,t (numpy.ndarray, numpy.array): Rigid transformation represented as
        a rotation matrix and translation.
    """
    s1 = np.sin(theta_z)
    s2 = np.sin(theta_y) 
    s3 = np.sin(theta_x)
    c1 = np.cos(theta_z) 
    c2 = np.cos(theta_y) 
    c3 = np.cos(theta_x)
    return np.array([[c1*c2, c1*s2*s3-s1*c3, c1*s2*c3+s1*s3],
                     [s1*c2, s1*s2*s3+c1*c3, s1*s2*c3-c1*s3],
                     [-s2,   c2*s3,          c2*c3]]), \
           np.array([dx, dy, dz])


def create_random_points(x_range, y_range, z_range, num_points=1):
    """
    Create one or more points uniformly distributed in the given range.
    
    Args:
        x_range (float): A list like structure so that max/min(x_range) returns 
            a float.  
        y_range (float): A list like structure so that max/min(y_range) returns 
            a float.  
        z_range (float): A list like structure so that max/min(z_range) returns 
            a float.
        numPoints (int): Number of points to create, default is 1.
    
    Returns:
        A list of numpy 3D arrays, each entry represents a 3D point.
    """
    if num_points<1:
        raise ValueError("Number of points must be greater/equal 1.")
        #get the bounds on the x,y,z coordinates
    x_rng = [min(x_range), max(x_range)]
    y_rng = [min(y_range), max(y_range)]
    z_rng = [min(z_range), max(z_range)]
    data = np.random.rand(3, num_points)
    data[0] = data[0]*(x_rng[1]-x_rng[0])+x_rng[0]
    data[1] = data[1]*(y_rng[1]-y_rng[0])+y_rng[0]    
    data[2] = data[2]*(z_rng[1]-z_rng[0])+z_rng[0]
    return [np.array(p) for p in np.transpose(data).tolist()]      
  

def transform_points(R, t, points):
    """
    Apply the given transformation to the points.
    
    Args:
    R (numpy 3x3 array): A rotation matrix (we do not check for orthonormality
                         so it is up to the caller to ensure that this is a 
                         rotation).
    t (numpy 4x4 array): A rigid tranfsormation represented by a homogenous
        transformation.
    points (list(numpy array)): List of 3D points.
    
    Returns:
        A list of numpy 3D arrays, each entry represents a 3D point.    
    """ 
    return [R.dot(np.array(p)) + t for p in points]
 
        
def transform_points_adding_noise(R, t, points, error_bound):
    """
    Apply the given transformation to the points and add noise to each coordinate.
    
    Args:
    R (numpy 3x3 array): A rotation matrix (we do not check for orthonormality
                         so it is up to the caller to ensure that this is a 
                         rotation).
    t (numpy 4x4 array): A rigid tranfsormation represented by a homogenous
        transformation.
    points (list(numpy array)): List of 3D points.
    error_bound (float): The noise added to each coordinate is in 
        (-error_bound, error_bound).
    
    Returns:
        A list of numpy 3D arrays, each entry represents a 3D point.    
    """ 
    
    return [R.dot(np.array(p)) + t + 
            np.random.uniform(-error_bound, error_bound, 3) for p in points]

if __name__ == "__main__":
    
    #change the numpy sring representation of arrays so that it displays
    #two digits, including trailing zeros, after the decimal point and 
    #suppresses scientific notation for small numbers
    original_print_options = np.get_printoptions()
    np.set_printoptions(formatter={'float': '{: 0.2f}'.format}, suppress=True)
    
    #create an arbitrary rigid transformation using the following parameters
    R,t=create_fixed_axis_frameD(theta_x=24.0, theta_y=45.0, theta_z=10.0, 
                                 dx=10.0, dy=2.0, dz=567.0)    
    print('The known rotation is:\n{0}\n'.format(R))
    print('The known translation is:\n{0}\n'.format(t))
    
    #randomly create points in left coordinate system inside the bounds
    points_in_left = create_random_points(x_range = (-100,100), 
                                          y_range = (-50,50),
                                          z_range = (-25,25),
                                          num_points = 5)
    #compute transformation in ideal - no noise settings
    points_in_right = transform_points(R, t, points_in_left)
    
    R,t = compute_pose_quaternion(points_in_left, points_in_right)
    print('No noise (quaternion):\n')
    print('The estimated rotation is:\n{0}\n'.format(R))
    print('The estimated translation is:\n{0}\n'.format(t))
    R,t = compute_pose_matrix(points_in_left, points_in_right)
    print('No noise (matrix):\n')
    print('The estimated rotation is:\n{0}\n'.format(R))
    print('The estimated translation is:\n{0}\n'.format(t))
    
    #now repeat with noise
    points_in_right = transform_points_adding_noise(R, t, points_in_left, error_bound=3.0)

    R,t = compute_pose_quaternion(points_in_left, points_in_right)
    print('With noise (quaternion):\n')
    print('The estimated rotation is:\n{0}\n'.format(R))
    print('The estimated translation is:\n{0}\n'.format(t))
    R,t = compute_pose_matrix(points_in_left, points_in_right)
    print('With noise (matrix):\n')
    print('The estimated rotation is:\n{0}\n'.format(R))
    print('The estimated translation is:\n{0}\n'.format(t))
        
    #return the original numpy print settings    
    np.set_printoptions(**original_print_options)    
                                        
